@@ -25,14 +25,28 @@
 package Image::ExifTool::Photoshop;
 
 use strict;
-use vars qw($VERSION $AUTOLOAD $iptcDigestInfo);
+use vars qw($VERSION $AUTOLOAD $iptcDigestInfo %printFlags);
 use Image::ExifTool qw(:DataAccess :Utils);
 
-$VERSION = '1.65';
+$VERSION = '1.72';
 
 sub ProcessPhotoshop($$$);
 sub WritePhotoshop($$$);
 sub ProcessLayers($$$);
+sub ProcessChannelOptions($$$);
+
+# PrintFlags bit definitions (ref forum13785)
+%printFlags = (
+    0 => 'Labels',
+    1 => 'Corner crop marks',
+    2 => 'Color bars', # (deprecated)
+    3 => 'Registration marks',
+    4 => 'Negative',
+    5 => 'Emulsion down',
+    6 => 'Interpolate', # (deprecated)
+    7 => 'Description',
+    8 => 'Print flags',
+);
 
 # map of where information is stored in PSD image
 my %psdMap = (
@@ -57,11 +71,11 @@ my %thumbnailInfo = (
     Protected => 1,
     RawConv => 'my $img=substr($val,0x1c); $self->ValidateImage(\$img,$tag)',
     ValueConvInv => q{
-        my $et = new Image::ExifTool;
+        my $et = Image::ExifTool->new;
         my @tags = qw{ImageWidth ImageHeight FileType};
         my $info = $et->ImageInfo(\$val, @tags);
         my ($w, $h, $type) = @$info{@tags};
-        $w and $h and $type eq 'JPEG' or warn("Not a valid JPEG image\n"), return undef;
+        $w and $h and $type and $type eq 'JPEG' or warn("Not a valid JPEG image\n"), return undef;
         my $wbytes = int(($w * 24 + 31) / 32) * 4;
         return pack('N6n2', 1, $w, $h, $wbytes, $wbytes * $h, length($val), 24, 1) . $val;
     },
@@ -106,7 +120,17 @@ my %unicodeString = (
     0x03f0 => { Unknown => 1, Name => 'PStringCaption' },
     0x03f1 => { Unknown => 1, Name => 'BorderInformation' },
     0x03f2 => { Unknown => 1, Name => 'BackgroundColor' },
-    0x03f3 => { Unknown => 1, Name => 'PrintFlags', Format => 'int8u' },
+    0x03f3 => {
+        Unknown => 1,
+        Name => 'PrintFlags',
+        Format => 'int8u',
+        PrintConv => q{
+            my $byte = 0;
+            my @bits = $val =~ /\d+/g;
+            $byte = ($byte << 1) | ($_ ? 1 : 0) foreach reverse @bits;
+            return DecodeBits($byte, \%Image::ExifTool::Photoshop::printFlags);
+        },
+    },
     0x03f4 => { Unknown => 1, Name => 'BW_HalftoningInfo' },
     0x03f5 => { Unknown => 1, Name => 'ColorHalftoningInfo' },
     0x03f6 => { Unknown => 1, Name => 'DuotoneHalftoningInfo' },
@@ -299,7 +323,13 @@ my %unicodeString = (
     0x0432 => { Unknown => 1, Name => 'MeasurementScale' }, #7
     0x0433 => { Unknown => 1, Name => 'TimelineInfo' }, #7
     0x0434 => { Unknown => 1, Name => 'SheetDisclosure' }, #7
-    0x0435 => { Unknown => 1, Name => 'DisplayInfo' }, #7
+    0x0435 => {
+        Name => 'ChannelOptions', #7/forum16762
+        SubDirectory => {
+            TagTable => 'Image::ExifTool::Photoshop::ChannelOptions',
+            Start => 4,
+        },
+    },
     0x0436 => { Unknown => 1, Name => 'OnionSkins' }, #7
     0x0438 => { Unknown => 1, Name => 'CountInfo' }, #7
     0x043a => { Unknown => 1, Name => 'PrintInfo2' }, #7
@@ -327,11 +357,46 @@ my %unicodeString = (
     0x2710 => { Unknown => 1, Name => 'PrintFlagsInfo' },
 );
 
+# Photoshop channel options (ref forum16762)
+%Image::ExifTool::Photoshop::ChannelOptions = (
+    PROCESS_PROC => \&ProcessChannelOptions,
+    VARS => { IS_BINARY => 1 },
+    GROUPS => { 2 => 'Image' },
+    0 => {
+        Name => 'ChannelColorSpace',
+        Format => 'int16u',
+        PrintConv => {
+            0 => 'RGB',
+            1 => 'HSB',
+            2 => 'CMYK',
+            7 => 'Lab',
+            8 => 'Grayscale',
+        },
+    },
+    2 => {
+        Name => 'ChannelColorData',
+        Format => 'int16u[4]',
+    },
+    11 => {
+        Name => 'ChannelOpacity',
+        PrintConv => '"$val%"',
+    },
+    12 => {
+        Name => 'ChannelColorIndicates',
+        PrintConv => {
+            0 => 'Selected Areas',
+            1 => 'Masked Areas',
+            2 => 'Spot Color',
+        },
+    },
+);
+
 # Photoshop JPEG quality record (ref 2)
 %Image::ExifTool::Photoshop::JPEG_Quality = (
     PROCESS_PROC => \&Image::ExifTool::ProcessBinaryData,
     WRITE_PROC => \&Image::ExifTool::WriteBinaryData,
     CHECK_PROC => \&Image::ExifTool::CheckBinaryData,
+    DATAMEMBER => [ 1 ],
     FORMAT => 'int16s',
     GROUPS => { 2 => 'Image' },
     0 => {
@@ -342,6 +407,7 @@ my %unicodeString = (
     },
     1 => {
         Name => 'PhotoshopFormat',
+        RawConv => '$$self{PhotoshopFormat} = $val',
         PrintConv => {
             0x0000 => 'Standard',
             0x0001 => 'Optimized',
@@ -350,6 +416,7 @@ my %unicodeString = (
     },
     2 => {
         Name => 'ProgressiveScans',
+        Condition => '$$self{PhotoshopFormat} == 0x0101',
         PrintConv => {
             1 => '3 Scans',
             2 => '4 Scans',
@@ -545,6 +612,13 @@ my %unicodeString = (
         ValueConv => '100 * $val / 255',
         PrintConv => 'sprintf("%d%%",$val)',
     },
+    _xvis  => {
+        Name => 'LayerVisible',
+        Format => 'int8u',
+        List => 1,
+        ValueConv => '$val & 0x02',
+        PrintConv => { 0x02 => 'No', 0x00 => 'Yes' },
+    },
     # tags extracted from additional layer information (tag ID's are real)
     # - must be able to accommodate a blank entry to preserve the list ordering
     luni => {
@@ -563,6 +637,16 @@ my %unicodeString = (
         List => 1,
         Unknown => 1,
     },
+    lclr => {
+        Name => 'LayerColors',
+        Format => 'int16u',
+        Count => 1,
+        List => 1,
+        PrintConv => {
+            0=>'None',  1=>'Red',  2=>'Orange', 3=>'Yellow',
+            4=>'Green', 5=>'Blue', 6=>'Violet', 7=>'Gray',
+        },
+    },
     shmd => { # layer metadata (undocumented structure)
         # (for now, only extract layerTime.  May also contain "layerXMP" --
         #  it would be nice to decode this but I need a sample)
@@ -577,6 +661,13 @@ my %unicodeString = (
         ValueConv => 'length $val ? ConvertUnixTime($val,1) : ""',
         PrintConv => 'length $val ? $self->ConvertDateTime($val) : ""',
     },
+    lsct => {
+        Name => 'LayerSections',
+        Format => 'int32u',
+        Count => 1,
+        List => 1,
+        PrintConv => { 0 => 'Layer', 1 => 'Folder (open)', 2 => 'Folder (closed)', 3 => 'Divider' },
+    },
 );
 
 # tags extracted from ImageSourceData found in TIFF images (ref PH)
@@ -656,7 +747,7 @@ sub ProcessLayersAndMask($$$)
     local $_;
     my ($et, $dirInfo, $tagTablePtr) = @_;
     my $raf = $$dirInfo{RAF};
-    my $fileType = $$et{VALUE}{FileType};
+    my $fileType = $$et{FileType};
     my $data;
 
     return 0 unless $fileType eq 'PSD' or $fileType eq 'PSB';   # (no layer section in CS1 files)
@@ -705,6 +796,25 @@ sub ProcessLayersAndMask($$$)
     return $raf->Seek($end, 0) ? 1 : 0;
 }
 
+#------------------------------------------------------------------------------
+# Process Photoshop channel options (ref forum16762)
+# Inputs: 0) ExifTool ref, 1) DirInfo ref, 2) tag table ref
+# Returns: 1 on success
+sub ProcessChannelOptions($$$)
+{
+    my ($et, $dirInfo, $tagTablePtr) = @_;
+    my $end = $$dirInfo{DirStart}  + $$dirInfo{DirLen};
+    $$dirInfo{DirLen} = 13;
+    my $i;
+    for ($i=0; $$dirInfo{DirStart} + 13 <= $end; ++$i) {
+        $$et{SET_GROUP1} = "Channel$i";
+        $et->ProcessBinaryData($dirInfo, $tagTablePtr);
+        $$dirInfo{DirStart} += 13;
+    }
+    delete $$et{SET_GROUP1};
+    return 1;
+}
+
 #------------------------------------------------------------------------------
 # Process Photoshop layers (beginning with layer count)
 # Inputs: 0) ExifTool ref, 1) DirInfo ref, 2) tag table ref
@@ -722,17 +832,16 @@ sub ProcessLayers($$$)
     my $pos = 0;
     return 0 if $dirLen < 2;
     $raf->Read($buff, 2) == 2 or return 0;
-    my $num = Get16s(\$buff, 0);
+    my $num = Get16s(\$buff, 0);    # number of layers
     $num = -$num if $num < 0;       # (first channel is transparency data if negative)
     $et->VerboseDir('Layers', $num, $dirLen);
     $et->HandleTag($tagTablePtr, '_xcnt', $num, Start => $pos, Size => 2, %dinfo); # LayerCount
     my $oldIndent = $$et{INDENT};
     $$et{INDENT} .= '| ';
-
     $pos += 2;
     my $psb = $$et{IsPSB};  # is PSB format?
     my $psiz = $psb ? 8 : 4;
-    for ($i=0; $i<$num; ++$i) {
+    for ($i=0; $i<$num; ++$i) { # process each layer
         $et->VPrint(0, $oldIndent.'+ [Layer '.($i+1)." of $num]\n");
         last if $pos + 18 > $dirLen;
         $raf->Read($buff, 18) == 18 or last;
@@ -750,6 +859,7 @@ sub ProcessLayers($$$)
         $sig =~ /^(8BIM|MIB8)$/ or last;    # verify signature
         $et->HandleTag($tagTablePtr, '_xbnd', undef, Start => 4, Size => 4, %dinfo);
         $et->HandleTag($tagTablePtr, '_xopc', undef, Start => 8, Size => 1, %dinfo);
+        $et->HandleTag($tagTablePtr, '_xvis', undef, Start =>10, Size => 1, %dinfo);
         my $nxt = $pos + 16 + Get32u(\$buff, 12);
         $n = Get32u(\$buff, 16);        # get size of layer mask data
         $pos += 20 + $n;                # skip layer mask data
@@ -797,7 +907,7 @@ sub ProcessLayers($$$)
                 $raf->Read($buff, $n) == $n or last;
                 $dinfo{DataPos} = $pos;
                 while ($count{$tag} < $i) {
-                    $et->HandleTag($tagTablePtr, $tag, '');
+                    $et->HandleTag($tagTablePtr, $tag, $tag eq 'lsct' ? 0 : '');
                     ++$count{$tag};
                 }
                 $et->HandleTag($tagTablePtr, $tag, undef, Start => 0, Size => $n, %dinfo);
@@ -814,6 +924,13 @@ sub ProcessLayers($$$)
         }
         $pos = $nxt;
     }
+    # pad lists if necessary to have an entry for each layer
+    foreach (sort keys %count) {
+        while ($count{$_} < $num) {
+            $et->HandleTag($tagTablePtr, $_, $_ eq 'lsct' ? 0 : '');
+            ++$count{$_};
+        }
+    }
     $$et{INDENT} = $oldIndent;
     return 1;
 }
@@ -835,7 +952,7 @@ sub ProcessDocumentData($$$)
     unless ($raf) {
         my $dataPt = $$dirInfo{DataPt};
         my $start = $$dirInfo{DirStart} || 0;
-        $raf = new File::RandomAccess($dataPt);
+        $raf = File::RandomAccess->new($dataPt);
         $raf->Seek($start, 0) if $start;
         $dirLen = length $$dataPt - $start unless defined $dirLen;
         $et->VerboseDump($dataPt, Start => $start, Len => $dirLen, Base => $$dirInfo{Base});
@@ -993,7 +1110,7 @@ sub ProcessPhotoshop($$$)
     if ($$et{VALUE}{IPTCDigest} and $$et{VALUE}{CurrentIPTCDigest} and
         $$et{VALUE}{IPTCDigest} ne $$et{VALUE}{CurrentIPTCDigest})
     {
-        $et->WarnOnce('IPTCDigest is not current. XMP may be out of sync');
+        $et->Warn('IPTCDigest is not current. XMP may be out of sync');
     }
     delete $$et{LOW_PRIORITY_DIR}{'*'};
     return $success;
@@ -1144,7 +1261,7 @@ be preserved when copying Photoshop information via user-defined tags.
 
 =head1 AUTHOR
 
-Copyright 2003-2022, Phil Harvey (philharvey66 at gmail.com)
+Copyright 2003-2024, Phil Harvey (philharvey66 at gmail.com)
 
 This library is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
